from ...core.basejob import SingleJob
from .scmjob import SCMResults

__all__ = ['MOPACJob', 'MOPACResults']


class MOPACResults(SCMResults):
    """A class for results of computation done with MOPAC.

    This class inherits all methods from |SCMResults|.

    .. technical::

        In case of a MOPAC job, preparation is much different from other programs of AMSuite, but the result handling is quite similar due to presence of KF files. Therefore |MOPACResults| is a subclass of |SCMResults|, but |MOPACJob| is not a subclass of |SCMJob|.
    """
    _rename_map = {'results.rkf':'$JN.rkf', '$JN.in.aux':'$JN.aux', '$JN.in.arc':'$JN.arc', '$JN.in.out':'$JN.out' }
    _kfext = '.rkf'

    def _int2inp(self):
        return list(range(1, 1+len(self.job.molecule)))

    def _atomic_numbers_input_order(self):
        return self.readkf('Molecule', 'AtomicNumbers')


class MOPACJob(SingleJob):
    """A class representing a single computational job with MOPAC."""
    _result_type = MOPACResults
    _command = 'MOPAC2016-SCM.exe'

    def get_input(self):
        """Transform the contents of ``input`` branch of ``settings`` into the first line of MOPAC input. Print the molecular coordinates together with frozen coordinate flags."""
        aux = self.settings.input.find_case('aux')
        if aux not in self.settings.input:
            self.settings.input[aux] = [0,'PRECISION=9']

        keylist = []
        for key, value in self.settings.input.items():
            if value is True:
                keylist.append(key)
            elif isinstance(value, tuple):
                keylist.append(key + '=(' + ','.join(map(str, value)) + ')' )
            elif isinstance(value, list):
                keylist.append(key + '(' + ','.join(map(str, value)) + ')' )
            else:
                keylist.append(key + '=' + str(value))
        ret = ' '.join(keylist) + '\n\n\n'

        for at in self.molecule:
            line = '{:7}'.format(at.symbol)
            for c in ['x', 'y', 'z']:
                num = at.__getattribute__(c)
                frz = 0 if ('mopac_freeze' in at.properties and c in at.properties.mopac_freeze) else 1
                line += ' {: 11.8f} {:d}'.format(num,frz)
            ret += line + '\n'
        return ret


    def get_runscript(self):
        """Generate a MOPAC runscript.

        The name of the MOPAC executable is taken from class attribute ``MOPACJob._command``. If you experience problems running MOPAC, check if that value corresponds to the name of the executable and this executable is visible in your ``$PATH`` (in case of AMSuite it's in ``$AMSBIN``). Note that a bare MOPAC executable should be used here, please avoid using any wrappers.

        The execution of MOPAC binary is followed by calling a simple command line tool ``tokf`` which reads various output text files produced by MOPAC and collects all the data in a binary KF file. See :ref:`kf-files` for details.
        """
        ret = self._command + ' ' + self._filename('inp')
        if self.settings.runscript.stdout_redirect:
            ret += ' >'+self._filename('out')
        ret += '\n\n'
        ret += 'cp {} {}.stdout\n'.format(self._filename('err'), self._filename('inp'))
        ret += 'tokf mopac {} {}.rkf\n'.format(self._filename('inp'), self.name)
        ret += 'rm {}.stdout\n\n'.format(self._filename('inp'))
        ret += 'rm {}\n\n'.format(self._filename('out'))
        return ret


    def check(self):
        """Grep standard output for ``* JOB ENDED NORMALLY *``."""
        s = self.results.grep_output('* JOB ENDED NORMALLY *')
        return len(s) > 0



import numpy as np
from collections import OrderedDict
from itertools import combinations

from .atom import Atom
from .molecule import Molecule
from ..core.private import sha256
from ..core.functions import add_to_class
from ..tools.units import Units


__all__ = ['label_atoms']


possible_flags = ['BO', 'RS', 'EZ', 'DH', 'CO', 'H2']


def twist(v1, v2, v3, tolerance=None):
    """
    Given 3 vectors in 3D space measure their "chirality" with *tolerance*.

    Returns a pair. The first element is an integer number measuring the orientation (clockwise vs counterclockwise) of *v1* and *v3* while looking along *v2*. Values 1 and -1 indicate this case and the second element of returned pair is ``None``. Value 0 indicates that *v1*, *v2*, and *v3* are coplanar, and the second element of the returned pair is indicating if two turns made by going *v1*->*v2*->*v3* are the same (left-left, right-right) or the opposite (left-right, right-left).
    """
    tolerance = 1e-2 if tolerance is None else tolerance
    v1 /= np.linalg.norm(v1)
    v2 /= np.linalg.norm(v2)
    v3 /= np.linalg.norm(v3)
    x = np.dot(v3, np.cross(v1,v2))
    if abs(x) <= tolerance:  #v1, v2, v3 are coplanar
        return 0, int(np.sign(np.dot(np.cross(v1,v2), np.cross(v2,v3))))
    return int(np.sign(x)), None


def bend(v1, v2, tolerance=None):
    """Check if two vectors in 3D space are parallel or perpendicular, with *tolerance* (in degrees).

    Returns 1 if *v1* and *v2* are collinear, 2 if they are perpendicular, 0 otherwise."""
    tolerance = 7.5 if tolerance is None else tolerance
    v1 /= np.linalg.norm(v1)
    v2 /= np.linalg.norm(v2)
    angle = Units.convert(abs(np.arccos(np.dot(v1,v2))), 'rad', 'deg')
    if angle <= tolerance:
        return 1
    if abs(angle - 90.0) <= tolerance:
        return 2
    return 0


def unique_atoms(atomlist):
    """Filter *atomlist* (list or |Molecule|) for atoms with unique ``IDname``."""
    d = {}
    for atom in atomlist:
        if atom.IDname not in d:
            d[atom.IDname] = 0
        d[atom.IDname] += 1
    return [atom for atom in atomlist if d[atom.IDname] == 1]


def initialize(molecule):
    """Initialize atom labeling algorithm by setting ``IDname`` and ``IDdone`` attributes for all atoms in *molecule*."""
    for at in molecule:
        at.IDname = at.symbol
        at.IDdone = False


def clear(molecule):
    """Remove ``IDname`` and ``IDdone`` attributes from all atoms in *molecule*."""
    for at in molecule:
        if hasattr(at, 'IDname'): del at.IDname
        if hasattr(at, 'IDdone'): del at.IDdone


def iterate(molecule, flags):
    """Perform one iteration of atom labeling alogrithm.

    First, mark all atoms that are unique and have only unique neighbors as "done". Then calculate new label for each atom that is not done. Return True if the number of different atom labels increased during this iteration.
    """
    names = len(set(at.IDname for at in molecule))
    unique = set(unique_atoms(molecule))

    for atom in molecule:
        if atom in unique and all(N in unique for N in atom.neighbors()):
            atom.IDdone = True
        if not atom.IDdone:
            atom.IDnew = new_name(atom, flags)

    for atom in molecule:
        if not atom.IDdone:
            atom.IDname = atom.IDnew

    new_names = len(set(atom.IDname for atom in molecule))
    return new_names > names #True means this iteration increased the number of distinct names


def new_name(atom, flags):
    """Compute new label for *atom*.

    The new label is based on the existing label of *atom*, labels of all its neighbors and (possibly) some additional conformational information. The labels of neighbors are not obtained directly by reading neighbor's ``IDname`` but rather by a process called "knocking". The *atom* knocks all its bonds. Each knocked bond returns an identifier describing the atom on the other end of the bond. The identifier is composed of knocked atom's ``IDname`` together with some additional information desribing the character of the bond and knocked atom's spatial environment. The exact behavior of this mechanism is adjusted by the contents of *flags* dictionary (see :func:`label_atoms` for details).
    """

    knocks = [knock(atom, bond, flags) for bond in atom.bonds]
    knocks.sort(key=lambda x: x[0])
    labels = set(i[0] for i in knocks) #types of neighbors

    more = []
    if flags['RS'] and len(knocks) == 4 and len(labels) == 4: #4 different neighbors
        v1 = atom.vector_to(knocks[0][1])
        v2 = atom.vector_to(knocks[1][1])
        v3 = atom.vector_to(knocks[2][1])
        more.append('RS'+str(twist(v1, v2, v3, flags.get('twist_tol'))))

    if flags['CO'] and len(knocks) >= 4:
        d = OrderedDict()
        for label, at in knocks:
            if label not in d:
                d[label] = []
            d[label].append(at)
        for label in d:
            if len(d[label]) > 1:
                angles = [bend(atom.vector_to(a), atom.vector_to(b), flags.get('bend_tol')) for a,b in combinations(d[label], 2)]
                angles.sort()
            else: #len(d[label]) == 1
                v1 = atom.vector_to(d[label][0])
                angles = []
                for k in d:
                    if k != label:
                        angles.append(sorted(bend(v1, atom.vector_to(a), flags.get('bend_tol')) for a in d[k]))
            more.append('CO'+str(angles))

    return sha256('|'.join([atom.IDname] + [i[0] for i in knocks] + more))


def knock(A, bond, flags):
    """Atom *A* knocks one of its bonds.

    *bond* has to be a bond formed by atom *A*. The other end of this bond (atom S) returns its description, consisting of its ``IDname`` plus, possibly, some additional information. If *BO* flag is set, the description includes the bond order of *bond*. If *EZ* flag is set, the description includes additional bit of information whenever E/Z isomerism is possible. If *DH* flag is set, the description includes additional information for all dihedrals A-S-N-F such that A is a unique neighbor of S and F is a unique neighbor of N.
    """

    S = bond.other_end(A)
    ret = S.IDname

    if flags['BO'] and bond.order != 1:
        ret += 'BO' + str(bond.order)

    if flags['H2']:
        S_nbors = S.neighbors()
        S_nbors.remove(A)
        if len(S_nbors) == 3 and len(unique_atoms(S_nbors)) == 3:
            S_nbors.sort(key=lambda x: x.IDname)
            v1, v2, v3 = [S.vector_to(i) for i in S_nbors]
            t = twist(v1, v2, v3, flags.get('twist_tol'))
            ret += 'H2' + str(t)

    if flags['EZ'] or flags['DH']:
        S_unique = unique_atoms(S.neighbors())
        if A in S_unique: #*A* is a unique neighbor of *S*
            S_unique.remove(A)
            S_unique.sort(key=lambda x: x.IDname)
            for b in S.bonds:
                N = b.other_end(S)
                if N in S_unique:
                    N_unique = unique_atoms(N.neighbors())
                    if S in N_unique:
                        N_unique.remove(S)
                    N_unique.sort(key=lambda x: x.IDname)
                    if N_unique:
                        F = N_unique[0]
                        v1 = A.vector_to(S)
                        v2 = S.vector_to(N)
                        v3 = N.vector_to(F)
                        t = twist(v1, v2, v3, flags.get('twist_tol'))
                        if flags['DH']:
                            ret += 'DH' + str(t)
                        elif flags['EZ'] and b.order == 2 and t[0] == 0:
                            #A-S=N-F are coplanar
                            ret += 'EZ' + str(t[1])
                        break

    return (ret, S)


def label_atoms(molecule, **kwargs):
    """Label atoms in *molecule*.

    Boolean keyword arguments:

    *   *BO* -- include bond orders
    *   *RS* -- include R/S stereoisomerism
    *   *EZ* -- include E/Z stereoisomerism
    *   *DH* -- include some dihedrals to detect alkane rotamers and syn-anti conformers in cycloalkanes
    *   *CO* -- include more spatial info to detect different conformation of coordination complexes (flat square, octaedr etc.)

    Numerical keyword arguments:

    *   *twist_tol* -- tolerance for :func:`twist` function
    *   *bend_tol* -- tolerance for :func:`bend` function

    Diherdals considered with *DH* are all the dihedrals A-B-C-D such that A is a unique neighbor or B and D is a unique neighbor of C.

    For atoms with 4 or more neighbors, the *CO* flag includes information about relative positions of equivalent/non-equivalent neighbors by checking if vectors from the central atom to the neighbors form 90 or 180 degrees angles.
    """
    initialize(molecule)
    while iterate(molecule, kwargs):
        pass
    return molecule


def molecule_name(molecule):
    """Compute the label of the whole *molecule* based on ``IDname`` attributes of all the atoms."""
    names = [atom.IDname for atom in molecule]
    names.sort()
    return sha256(' '.join(names))


@add_to_class(Molecule)
def label(self, level=1, keep_labels=False, flags=None):
    """Compute the label of this molecule using chosen *level* of detail.

    Possible levels are:

    *   **0**: does not pefrom any atom labeling, returns empirical formula (see :meth:`~scm.plams.mol.molecule.Molecule.get_formula`)
    *   **1**: only direct connectivity is considered, without bond orders (in other words, treats all the bonds as single bonds)
    *   **2**: use connectivity and bond orders
    *   **3**: use connectivity, bond orders and some spatial information to distinguish R/S and E/Z isomers
    *   **4**: use all above, plus more spatial information to distinguish different rotamers and different types of coordination complexes

    If you need more precise control of what is taken into account while computing the label (or adjust the tolerance for geometrical operations) you can use the *flags* argument. It should be a dictionary of parameters recognized by :func:`~scm.plams.mol.identify.label_atoms`. Each of two letter boolean flags has to be present in *flags*. If you use *flags*, *level* is ignored.

    The *level* argument can also be a tuple of integers. In that case the labeling algorithm is run multiple times and the returned value is a tuple (with the same length as *level*) containing labels calculated with given levels of detail.

    This function, by default, erases ``IDname`` attributes of all atoms at the end. You can change this behavior with *keep_labels* argument.

    If the molecule does not contain bonds, :meth:`~scm.plams.mol.molecule.Molecule.guess_bonds` is used to determine them.


    .. note::

        This method is a new PLAMS feature and its still somewhat experimental. The exact details of the algorithm can, and probably will, change in future. You are more than welcome to provide any feedback or feature requests.

    """
    if isinstance(level, (tuple, list)):
        return tuple(self.label(i) for i in level)

    if flags is None:
        if level == 0:
            return self.get_formula()

        flags = {i:False for i in possible_flags}
        if level >= 2:
            flags['BO'] = True
        if level >= 3:
            flags['RS'] = True
            flags['EZ'] = True
        if level >= 4:
            flags['DH'] = True
            flags['CO'] = True

    if len(self.bonds) == 0:
        self.guess_bonds()

    clear(self)
    label_atoms(self, **flags)
    ret = molecule_name(self)
    if not keep_labels:
        clear(self)
    return ret

@add_to_class(Molecule)
def set_local_labels(self, niter=2, flags=None) :
    """
    Set atomic labels (IDnames) that are unique for local structures of a molecule

    * ``niter`` -- The number of iterations in the atom labeling scheme

    The idea of this method is that the number of iterations can be specified.
    If kept low (default niter), local structures over different molecules will have the same label.
    """
    if flags is None :
        flags = {i:False for i in possible_flags}
    initialize(self)
    for i in range(niter) :
        iterate(self,flags)

def find_permutation(molecule1, molecule2):
    """


    """
    if molecule1.label(0) != molecule2.label(0):
        return None





@add_to_class(Molecule)
def reorder(self, other):
    """Reorder atoms in this molecule to match the order in some *other* molecule. The reordering is applied only if the perfect match is found. Returned value is the applied permutation (as a list of integers) or ``None``, if no reordering was performed. See also :func:`~scm.plams.mol.identify.find_permutation`.
    """

    perm = find_permutation(self, other)
    if perm:
        self.atoms = [self.atoms[i] for i in perm]
        return perm
    return None




#!/usr/bin/env python

import numpy
from ..tools.periodic_table import PT
from ..mol.molecule import Molecule
from ..mol.atom import Atom
from ..core.settings import Settings
from ..core.errors import PlamsError
from .rkffile import RKFTrajectoryFile
from .rkffile import bohr_to_angstrom

__all__ = ['RKFHistoryFile']

class RKFHistoryFile (RKFTrajectoryFile) :
        """
        Class representing an RKF file containing a molecular simulation history with varying numbers of atoms

        An instance of this class has the following attributes:

        *   ``file_object`` -- A PLAMS |KFFile| object, referring to the actual RKF file
        *   ``position``    -- The frame to which the cursor is currently pointing in the RKF file
        *   ``mode``        -- Designates whether the file is in read or write mode ('rb' or 'wb')
        *   ``elements``    -- The elements of the atoms in the system at the current frame
        *   ``conect``      -- The connectivity information of the current frame
        *   ``mddata``      -- Read mode only: A dictionary containing data from the MDHistory section in the RKF file
        *   ``read_lattice``-- Read mode only: Wether the lattice vectors will be read from the file
        *   ``read_bonds``  -- Wether the connectivity information will be read from the file

        An |RKFHistoryFile| object behaves very similar to a regular file object.
        It has read and write methods (:meth:`read_next` and :meth:`write_next`) 
        that read and write from/to the position of the cursor in the ``file_object`` attribute. 
        If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
        the cursor to any frame in the file and reads from there.
        The amount of information stored in memory is kept to a minimum, as only information from the latest frame
        is ever stored.

        Reading and writing to and from the files can be done as follows::

            >>> from scm.plams import RKFHistoryFile

            >>> rkf = RKFHistoryFile('ams.rkf')
            >>> mol = rkf.get_plamsmol()

            >>> rkfout = RKFHistoryFile('new.rkf',mode='wb')

            >>> for i in range(rkf.get_length()) :
            >>>     crd,cell = rkf.read_frame(i,molecule=mol)
            >>>     rkfout.write_next(molecule=mol)
            >>> rkfout.close()

        The above script reads information from the RKF file ``ams.rkf`` into the |Molecule| object ``mol``
        in a step-by-step manner..
        The |Molecule| object is then passed to the :meth:`write_next` method of the new |RKFHistoryFile|
        object corresponding to the new rkf file ``new.rkf``.

        The exact same result can also be achieved by iterating over the instance as a callable

            >>> rkf = RKFHistoryFile('ams.rkf')
            >>> mol = rkf.get_plamsmol()

            >>> rkfout = RKFHistoryFile('new.rkf',mode='wb')

            >>> for crd,cell in rkf(mol) :
            >>>     rkfout.write_next(molecule=mol)
            >>> rkfout.close()

        This procedure requires all coordinate information to be passed to and from the |Molecule| object
        for each frame, which can be time-consuming.
        Some time can be saved by bypassing the |Molecule| object::

            >>> rkf = RKFHistoryFile('ams.rkf')

            >>> rkfout = RKFHistoryFile('new.rkf',mode='wb')

            >>> for crd,cell in rkf :
            >>>     rkfout.write_next(coords=crd,cell=cell,elements=rkf.elements,conect=rkf.conect)
            >>> rkfout.close()

        The only mandatory argument to the :meth:`write_next` method is ``coords``.
        Further time can be saved by setting the ``read_lattice`` and ``read_bonds`` variables to False.

        By default the write mode will create a minimal version of the RKF file, containing only elements,
        coordinates, lattice, and connectivity information.
        This minimal file format can be read by AMSMovie.

        If the original RKF file contains an MDHistory section (if it resulted from a MolecularGun simulation)
        it is possible to store the information from that section and write it to another file.
        To enable this, the method :meth:`store_mddata` needs to be called after creation,
        and a dictionary of mddata needs to be passed to the :meth:`write_next` method.
        When that is done, the AMS trajectory analysis tools can be used on the file.
        Restarting an MD run with such a file is however currently not possible::

            >>> rkf = RKFHistoryFile('ams.rkf')
            >>> rkf.store_mddata()
            >>> mol = rkf.get_plamsmol()

            >>> rkf_out = RKFHistoryFile('new.rkf',mode='wb')
            >>> rkf_out.store_mddata(rkf)

            >>> for i in range(rkf.get_length()) :
            >>>         crd,cell = rkf.read_frame(i,molecule=mol)
            >>>         rkf_out.write_next(molecule=mol,mddata=rkf.mddata)
            >>> rkf_out.close()
        """

        def __init__ (self, filename, mode='rb', fileobject=None, ntap=None) :
                """
                Initializes the RKFHistoryFile object

                * ``filename``   -- The path to the RKF file
                * ``mode``       -- The mode in which to open the RKF file ('rb' or 'wb')
                * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
                * ``ntap``       -- If the file is in write mode, the number of atoms can be passed here
                """
                self.added_atoms = None
                self.removed_atoms = None
                self.chemical_systems = None
                RKFTrajectoryFile.__init__(self,filename,mode,fileobject,ntap)

                self.input_elements = self.elements[:]
                self.versionhistory_length = 0
                self.system_version_elements = {}
                self.frame = -1
                self.props = None

                # This is for writing only
                self.system_version_props = {}
                self.version_history_items = []

        def set_elements (self, elements) :
                """
                Sets the elements attribute (needed in write mode).

                *   ``elements`` -- A list containing the element name of each atom
                """
                if self.position > 0 :
                        raise PLAMSError('Elements should not be changed while reading/writing is already in progress')
                RKFTrajectoryFile.set_elements(self, elements)
                self.input_elements = elements

        def get_plamsmol (self) :
                """
                Extracts a PLAMS molecule object from the RKF file
                """
                section_dict = self.file_object.read_section('ChemicalSystem(1)')
                if len(section_dict) == 0 :
                        section_dict = self.file_object.read_section('InputMolecule')
                if len(section_dict) == 0 :
                        section_dict = self.file_object.read_section('Molecule')
                plamsmol = Molecule._mol_from_rkf_section(section_dict)
                return plamsmol

        def _rewrite_molecule (self) :
                """
                Overwrite the molecule section with the latest frame (called in close())
                """
                mol = self.get_plamsmol()
                crd,cell = self.read_last_frame(molecule=mol)
                self._write_molecule_section(crd,cell,molecule=mol)

        def _read_header (self) :
                """
                Read the start molecule data from the InputMolecule section (not the Molecule section)
                """
                self.added_atoms = {}
                self.removed_atoms = {} 
                self.chemical_systems = {0:1} 
                secname = 'ChemicalSystem(1)'
                if self.file_object.reader._sections is None :
                        self.file_object.reader._create_index()
                if not 'SystemVersionHistory' in self.file_object.reader._sections :
                        secname = 'InputMolecule'
                RKFTrajectoryFile._read_header (self, molecule_section=secname)

                # Now store the added and removed atoms along the trajectory
                # (This might be slow?)
                # I could also do it on the fly, but that may be messy when we move back and forth through the file
                if not 'SystemVersionHistory' in self.file_object.reader._sections :
                        return
                #version = 0
                version = 1
                self._set_system_version_elements()
                for i in range(self.get_length()) :
                        new_version = self.file_object.read('History','SystemVersion(%i)'%(i+1))
                        if new_version == version :
                                continue
                        self.added_atoms[i] = {}
                        self.removed_atoms[i] = {}
                        # Now look for the added and removed atoms
                        removed_atoms = []
                        if 'RemovedAtoms(%i)'%(new_version) in self.file_object.reader._sections['SystemVersionHistory'] :
                                removed_atoms = self.file_object.read('SystemVersionHistory','RemovedAtoms(%i)'%(new_version))
                        if not isinstance(removed_atoms,list) :
                                removed_atoms = [removed_atoms]
                        added_atoms = []
                        if 'AddedAtoms(%i)'%(new_version) in self.file_object.reader._sections['SystemVersionHistory'] :
                                added_atoms = self.file_object.read('SystemVersionHistory','AddedAtoms(%i)'%(new_version))
                        if not isinstance(added_atoms,list) :
                                added_atoms = [added_atoms]
                        # Now find the corresponding elements
                        chemSysNum = self.file_object.read('SystemVersionHistory','SectionNum(%i)'%(new_version))
                        # Compare to the previous chemical system
                        #prev_version = new_version - 1
                        #prevChemSysNum = self.file_object.read('SystemVersionHistory','SectionNum(%i)'%(prev_version))
                        prevChemSysNum = self.chemical_systems[max(self.chemical_systems.keys())]
                        sectionname = 'ChemicalSystem(%i)'%(prevChemSysNum)
                        prev_elements = [PT.get_symbol(atnum) for atnum in self.file_object.read(sectionname,'AtomicNumbers')]
                        sectionname = 'ChemicalSystem(%i)'%(chemSysNum)
                        elements = [PT.get_symbol(atnum) for atnum in self.file_object.read(sectionname,'AtomicNumbers')]
                        ################
                        # Read badly written files
                        diff = (len(added_atoms) - len(removed_atoms)) - (len(elements)-len(prev_elements))
                        if diff != 0 :
                                chemSysNum, elements = self._correct_chemical_system(elements,prev_elements,added_atoms,removed_atoms)
                        #print ('chemSysNum: ',i,chemSysNum, len(elements),added_atoms,removed_atoms)
                        ####################
                        removed_elements = [prev_elements[i-1].split('.')[0] for i in removed_atoms]
                        added_elements = [elements[i-1].split('.')[0] for i in added_atoms]
                        # Now store the elements
                        for iat,el in zip(removed_atoms,removed_elements) :
                                self.removed_atoms[i][iat] = el 
                        for iat,el in zip(added_atoms,added_elements) :
                                self.added_atoms[i][iat] = el
                        self.chemical_systems[i] = chemSysNum
                        version = new_version

        def _correct_chemical_system (self, elements, prev_elements, added_atoms, removed_atoms) :
                """
                Check if the referenced chemical system is correct, and if not, find one matching added/removed atoms
                """
                #print ('Searching for chemical system elsewhere')
                new_elements = [el for el in prev_elements]
                for ra in reversed(removed_atoms) :
                        new_elements.pop(ra-1)
                for aa in added_atoms :
                        new_elements.insert(aa-1,'*') # Add a wildcard
                chemSysNum = self._check_for_chemical_system(new_elements)
                if chemSysNum is None :
                        print ('Chemical system does not exist')
                        chemSysNum = self._check_for_chemical_system(new_elements,compare_elements=False)
                        if chemSysNum is None :
                                raise Exception('Chemical system with this number of atoms does not exist')
                sectionname = 'ChemicalSystem(%i)'%(chemSysNum)
                elements = [PT.get_symbol(atnum) for atnum in self.file_object.read(sectionname,'AtomicNumbers')]
                ###################
                # Check the regions
                section_dict = self.file_object.read_section(sectionname)
                plamsmol = Molecule._mol_from_rkf_section(section_dict)
                print (0,plamsmol.atoms[0].properties.suffix)
                for aa in added_atoms :
                        print (aa-1,plamsmol.atoms[aa-1].properties.suffix)
                ###################
                return chemSysNum, elements

        # FIXME: The _write_header section writes the starting molecule to the Molecule section, 
        #        not the final molecule (like the Fortran code does)
        def _write_header (self, coords, cell, molecule=None) :
                """
                Write Molecule info to file (elements, periodicity)
                """
                # First write the general section
                self._write_general_section()
                
                # Then write the input molecule
                self._update_celldata(cell)
                self._write_molecule_section(coords, cell, molecule=molecule)
                # I think the InputMolecule is mandatory in this case
                self._write_molecule_section(coords, cell, section='InputMolecule', molecule=molecule)
                self.chemical_systems = {}
                if self.include_mddata :
                        # Start setting up the MDHistory section as well
                        self.file_object.write('MDHistory','blockSize',100)

                self.added_atoms = {}
                self.removed_atoms = {}

        def _read_coordinates (self, i, molecule, cell, bonds) :
                """
                Read the coordinates at step i
                """
                coords = self.coords.reshape(len(self.coords)*3)

                update_molecule = False
                elements = self._read_elements_for_frame(i)
                if elements != self.elements :
                        update_molecule = True
                if isinstance(molecule,Molecule) :
                        _, _, mol_elements, _, props = self._read_plamsmol(molecule)
                        if mol_elements != self.elements or props != self.props : # This molecule has nothing to do with the previously read one
                                update_molecule = True
                        prev_frames = [iframe for iframe in self.chemical_systems.keys() if iframe<=i]
                        ifr = 0
                        if len(prev_frames) > 0 :
                                ifr = prev_frames[-1]
                        next_frames = [iframe for iframe in self.chemical_systems.keys() if iframe>i]
                        jfr = len(self)
                        if len(next_frames) > 0 :
                                jfr = next_frames[0]
                        # If the previous frame falls before the last chemical system or after the next one, update
                        if self.frame < ifr or self.frame >= jfr :
                                update_molecule = True

                if update_molecule :
                        self.elements = elements
                        self.frame = i
                        self.coords = numpy.zeros((len(elements),3))
                        coords = self.coords.reshape((len(elements)*3))
                        # Rebuild the molecule (bonds will disappear for now)
                        if isinstance(molecule,Molecule) :
                                self.props = props
                                secname = 'ChemicalSystem(%i)'%(self.chemical_systems[ifr])
                                section_dict = self.file_object.read_section(secname)
                                new_mol = Molecule._mol_from_rkf_section(section_dict)
                                for at in reversed(molecule.atoms) :
                                        molecule.delete_atom(at)
                                molecule.properties = Settings()
                                for iel,el in enumerate(elements) :
                                        atom = new_mol.atoms[iel]
                                        #atom = Atom(PT.get_atomic_number(el))
                                        molecule.add_atom(atom)
                                # Now what if the elements found in this ChemicalSystem section
                                # do not match the expected elements (self.elements)?
                                new_elements = [at.symbol for at in molecule.atoms]
                                if new_elements != self.elements :
                                        print ('Chemical section does not match',i)
                                        atnums = [PT.get_atomic_number(el) for el in self.elements]
                                        for iat,atnum in enumerate(atnums) :
                                                molecule.atoms[iat].atnum = atnum

                coords[:] = self.file_object.read('History', 'Coords(%i)'%(i+1))
                coords *= bohr_to_angstrom
                # This changes self.coords behind the scenes

                # Assign the data to the molecule object
                if isinstance(molecule,Molecule) :
                        cell_reduced = None
                        if cell is not None :
                                cell_reduced = cell[:self.nvecs]
                        self._set_plamsmol(self.coords,cell_reduced,molecule,bonds)

        def _read_elements_for_frame (self, frame) :
                """
                Use the added and removed atoms to read the elements at each frame
                """
                elements = self.input_elements[:]
                for i in range(frame+1) :
                        if i in self.removed_atoms :
                                # Insert them into the elements list, in the correct place
                                elements = [el for iat,el in enumerate(elements) if not iat+1 in self.removed_atoms[i]]
                        if i in self.added_atoms :
                                for iat,el in self.added_atoms[i].items() :
                                        elements.insert(iat-1,el)
                return elements

        def write_next (self, coords=None, molecule=None, elements=None, cell=[0.,0.,0.], conect=None, historydata=None, mddata=None) :
                """
                Write frame to next position in trajectory file

                * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
                * ``molecule`` -- A molecule object to read the molecular data from
                * ``elements`` -- The element symbols of the atoms in the system
                * ``cell``     -- A set of lattice vectors (or cell diameters for an orthorhombic system)
                * ``conect``   -- A dictionary containing the connectivity info (e.g. {1:[2],2:[1]})
                * ``historydata`` -- A dictionary containing additional variables to be written to the History section
                * ``mddata``   -- A dictionary containing the variables to be written to the MDHistory section

                The ``mddata`` dictionary can contain the following keys:
                ('TotalEnergy', 'PotentialEnergy', 'Step', 'Velocities', 'KineticEnergy', 
                'Charges', 'ConservedEnergy', 'Time', 'Temperature')

                The ``historydata`` dictionary can contain for example:
                ('Energy','Gradients','StressTensor')
                All values must be in atomic units
                Numpy arrays or lists of lists will be flattened before they are written to the file

                .. note::

                        Either ``coords`` and ``elements`` or ``molecule`` are mandatory arguments
                """
                # Check for common error in the arguments
                if coords is not None :
                        if isinstance(coords,Molecule) :
                                raise PlamsError('The PLAMS molecule needs to be passed as the second argument (molecule)')

                props = None
                if isinstance(molecule,Molecule) :
                        coords, cell, elements, conect, props = self._read_plamsmol(molecule)
                # Make sure that the cell consists of three vectors
                cell = self._convert_cell(cell)
                if conect is not None :
                        if len(conect) == 0 : conect = None
                self.conect = conect

                # If this is the first step, write the header
                if self.position == 0 :
                        self.elements = elements
                        self.props = props
                        self._write_header(coords,cell,molecule=molecule)
                        # This is specific for the history-file object
                        self.input_elements = elements[:]

                # Define some local variables
                counter = 1
                step = self.position
                if mddata is not None :
                        if 'Step' in mddata :
                                step = mddata['Step']
                # Energy should be read from mddata first, otherwise from historydata, otherwise set to zero
                energy = self._set_energy(mddata, historydata)
                if not self.include_historydata :
                        historydata = {}
                historydata['Energy'] = energy

                # Write the history section
                counter = self._write_history_entry(step, coords, cell, conect, historydata)

                if self.include_mddata and mddata is not None :
                        self._write_mdhistory_entry(mddata)
                
                # If a change took place, write it.
                # Note: The chemical system is only written if the elements change. Changes in the atom properties are not checked
                if elements != self.elements or props != self.props or self.position==0 :
                        # version history has to be checked before the elements can be updated
                        chemsysversion = self._check_for_chemical_system(elements, props=props)
                        self._write_version_history(elements, coords, cell, props)
                        # Now update the elements and props
                        self.elements = elements
                        self.props = props
                        # Write the molecule sections
                        if chemsysversion is None :
                                chemsysversion = len(self.system_version_elements)
                                self._write_molecule_section (coords, cell, section='ChemicalSystem(%i)'%(chemsysversion), molecule=molecule)
                        self.chemical_systems[self.position] = chemsysversion
                counter = self._write_system_version_history_entry(counter)

                self.position += 1

                if self.saving_freq is not None :
                        if self.position%self.saving_freq == 0 : self.file_object.save()

        def _write_version_history (self, elements, coords, cell, props) :
                """
                Write the version history
                """
                # Enter the correct SystemVersionHistory entry into the History section
                self.versionhistory_length += 1
                version = self.versionhistory_length

                # Find the corresponding ChemicalSystem and write it
                chemsysversion = self._check_for_chemical_system(elements, props=props)
                if chemsysversion is None :
                        #self.system_version_elements.append(elements[:])
                        if len(self.system_version_elements) in self.system_version_elements.keys() :
                                raise Exception('self.system_version_elements is not consecutively numbered')
                        self.system_version_elements[len(self.system_version_elements)] = elements[:]
                        self.system_version_props[len(self.system_version_elements)] = props
                        chemsysversion = len(self.system_version_elements)

                # Now add an entry to SystemVersionHistory
                added_atoms, removed_atoms = self._find_system_change(elements, props)
                self.file_object.write('SystemVersionHistory','nEntries',version)
                self.file_object.write('SystemVersionHistory','currentEntryOpen',False)
                if 'SectionNum' not in self.version_history_items :
                        self.version_history_items.append('SectionNum')
                svh_counter = self.version_history_items.index('SectionNum') + 1
                self._write_keydata_in_history('SectionNum',svh_counter,False,1,version,chemsysversion,'SystemVersionHistory')
                # Avoid writing empty data into this section by storing the items.
                if len(added_atoms) > 0 :
                        if 'AddedAtoms' not in self.version_history_items :
                                self.version_history_items.append('AddedAtoms')
                        svh_counter = self.version_history_items.index('AddedAtoms') + 1
                        data = [iat+1 for iat in added_atoms]
                        self._write_keydata_in_history('AddedAtoms',svh_counter,False,len(data),version,data,'SystemVersionHistory')
                if len(removed_atoms) > 0 :
                        if 'RemovedAtoms' not in self.version_history_items :
                                self.version_history_items.append('RemovedAtoms')
                        svh_counter = self.version_history_items.index('RemovedAtoms') + 1
                        data = [iat+1 for iat in removed_atoms]
                        self._write_keydata_in_history('RemovedAtoms',svh_counter,False,len(data),version,data,'SystemVersionHistory')

        def _write_system_version_history_entry (self, counter) :
                """
                Write the entry for the SystemVersionHistory into the History section
                """
                version = self.versionhistory_length
                self._write_keydata_in_history('SystemVersion', counter, False, 1, self.position+1, version)
                counter += 1
                return counter

        def _find_system_change (self, elements, props) :
                """
                Find out which atoms were added and/or deleted
                """
                # First find out which elements were removed
                removed_atoms = []
                position = 0
                for i,el in enumerate(self.elements) :
                        if position == len(elements) : 
                                removed_atoms += [j for j in range(i,len(self.elements))]
                                break
                        # Get the props, if relevant
                        if self.props is None : 
                                op = None
                        else :
                                op = self.props[i]
                        if props is None : 
                                p = None
                        else :
                                p = props[position]
                        # Find the changes
                        if el == elements[position] and op==p :
                                position += 1
                        else :
                                removed_atoms.append(i)

                # Then find out which elements were added
                added_atoms = [i for i in range(position,len(elements))]
                
                # Now store them (this is actually not really necessary)
                if len(removed_atoms) > 0 :
                        self.removed_atoms[self.position] = {}
                if len(added_atoms) > 0 :
                        self.added_atoms[self.position] = {}
                for iat in removed_atoms :
                        self.removed_atoms[self.position][iat+1] = self.elements[iat]
                for iat in added_atoms :
                        self.added_atoms[self.position][iat+1] = elements[iat]
                return added_atoms, removed_atoms

        def _check_for_chemical_system (self, elements, compare_elements=True, props=None) :
                """
                Check if the new chemical system was encountered before
                """
                iterator = self.system_version_elements.items()

                version = None
                for ind,(i,prev_elements) in enumerate(iterator) :
                        equal = True
                        if len(prev_elements) != len(elements) :
                                continue
                        if compare_elements :
                                for el,pel in zip(elements,prev_elements) :
                                        if el != pel :
                                                if el == '*' : continue
                                                equal = False
                                                break
                        if props is not None and equal :
                                # This only occurs when writing, otherwise system_version_props is empty
                                if i in self.system_version_props.keys() :
                                        if props != self.system_version_props[i] :
                                                equal = False
                        if equal :
                                version = i+1
                return version

        def _set_system_version_elements (self) :
                """
                Store all chemical systems from the file
                """
                self.system_version_elements = {}
                keys = [key for key in self.file_object.reader._sections.keys() if 'ChemicalSystem' in key]
                nums = [int(k.split('(')[1].split(')')[0])-1 for k in keys]
                for num,key in zip(nums,keys) :
                        elements = [PT.get_symbol(atnum) for atnum in self.file_object.read(key,'AtomicNumbers')]
                        self.system_version_elements[num] = elements
